-- 2021-04-22 License CC0 (Creative Commons license) (c) darkfrei

function create_particle (data, id)
	local particle = {}
	particle.id = id
	particle.radius = data and data.radius or 50
	particle.mass = particle.radius*particle.radius
	particle.color = {
		0.5+0.5*(particle.radius%10)/10, 
		0.5+0.5*(particle.radius%20)/20, 
		0.5+0.5*(particle.radius%50)/50, 
	}
	particle.x = math.random (particle.radius, width - particle.radius)
	particle.y = math.random (particle.radius, height - particle.radius)
	local vel = 100*math.random ()
	local angle = math.random () * 2 * math.pi
	particle.vx =  vel * math.cos(angle)
	particle.vy = -vel * math.sin(angle)
	particle.ax, particle.ay = 0, 0
	table.insert(particles, particle)
end

function create_grid ()
	grid = {}
	local size = grid_size
	local max_x, max_y = width-size, height-size
	for x = 0, max_x, size do
		grid[x]={}
		for y = 0, max_y, size do
			local a = 132/255
			local b = 10/255
			local c = (x/size+y/size)%2
			grid[x][y]={color = {a+b*c, a+b*c, a+b*c}, size = size, particles={}}
--			print ('x: '..x..' y:'..y)
		end
	end
end

function love.load()
	love.graphics.setLineWidth( 2 )
	local ddwidth, ddheight = love.window.getDesktopDimensions( display )
	if ddheight > 1080 then -- for video capture
		print('ddheight: ' .. ddheight)
		love.window.setMode(1920, 1080, {resizable=false, borderless=true})
	else
		love.window.setMode(ddwidth, ddheight-200, {resizable=true, borderless=false})
	end
	
	grid_size = 50
	
	width, height = love.graphics.getDimensions( )
	width, height = math.floor(width/grid_size)*grid_size, math.floor(height/grid_size)*grid_size
	
	
	particles = {}
	create_grid ()
	for i = 1, 20 do
		create_particle ({radius = math.random (10, 100)}, i)
	end
	collidings = {}
end

function update_speed (particle, dt)
	local ddx, ddy = particle.ax*dt, particle.ay*dt
	particle.vx=particle.vx+ddx
	particle.vy=particle.vy+ddy
end

function update_position (particle, dt)
	local dx, dy = particle.vx*dt, particle.vy*dt
	particle.x=particle.x+dx
	particle.y=particle.y+dy
	
	if particle.x < particle.radius  then -- continuous collision detection
		particle.x = particle.x - (particle.radius-particle.x)/dx
		particle.vx = - particle.vx
	elseif particle.x > (width - particle.radius) then
		particle.x = particle.x + (width - particle.radius-particle.x)/dx
		particle.vx = - particle.vx
	end
	if particle.y < particle.radius then
		particle.y = particle.y - (particle.radius-particle.y)/dy
		particle.vy = - particle.vy
		elseif particle.y > (height - particle.radius) then
		particle.y = particle.y  + (height - particle.radius-particle.y)/dy
		particle.vy = - particle.vy
	end
end

function update_particle (particle, dt)
--	print('update speed')
	update_speed (particle, dt)
--	print('update position')
	update_position (particle, dt)
end
 

function get_variations (list)
	local variations = {}
	for i=1, #list-1 do
		for j=i+1, #list do
			table.insert (variations, {list[i], list[j]})
		end
	end
	return variations
end

function is_overlap (particles)
	if #particles == 2 then
		local distance = (particles[2].x-particles[1].x)^2+(particles[2].y-particles[1].y)^2
		local b = (particles[1].radius+particles[2].radius)^2
		local overlap = distance <= (b+0.000001)
		if overlap then
			local minid = math.min (particles[1].id, particles[2].id)
			local maxid = math.max (particles[1].id, particles[2].id)
			collisions[minid] = collisions[minid] or {}
			if not collisions[minid][maxid] then
				collisions[minid][maxid] = {
					particles ={particles[1], particles[2]},
					distance = distance
					}
			end
			return true
		end
		return false
	elseif #particles > 2 then
		local variations = get_variations (particles)
		local overlap = false
		for i, v in pairs (variations) do
			if is_overlap (v) then
				overlap = true
			end
		end
		return overlap
	end
end

--function pushback (particles)
--	if #particles == 2 then
--		local dx = particles[2].x - particles[1].x
--		local dy = particles[2].y - particles[1].y
--		local d = (dx*dx+dy*dy)^0.5
----		print ('d:'..d)
--		local overlap = - 0.5*(d - particles[1].radius - particles[2].radius)
----		print ('overlap:'..overlap)
		
--		if (overlap > 1) and (d > 1) then 
--			table.insert (collidings, {particles[1], particles[2]})
--			local odx = overlap*dx/d -- normalizetion
--			local ody = overlap*dy/d
--			particles[1].x = particles[1].x - odx
--			particles[2].x = particles[2].x + odx
--			particles[1].y = particles[1].y - ody
--			particles[2].y = particles[2].y + ody
--		end
--	elseif #particles > 2 then
--		local variations = get_variations (particles)
--		for i, v in pairs (variations) do
--			pushback (v)
--		end
--	end
--end

--function solve_grid (dt)
--	collidings = {}
	
--	for x,ys in pairs (grid) do
--		for y, g in pairs (ys) do
--			if #g.particles >= 2 then
--				g.near = true
--			else
--				g.near = false
--			end
--			if is_overlap (g.particles) then
--				g.overlap = true
----				pushback (g.particles)
--			else
--				g.overlap = false
--			end
			
			
--			-- erase all old data
----			g.particles = {}
--		end
--	end
--end


function update_grid (dt)
	-- erase all old data
	for x,ys in pairs (grid) do
		for y, g in pairs (ys) do
			g.particles = {}
		end
	end
	
	-- 
	for i, particle in pairs (particles) do
		local x1 = math.floor ((particle.x-particle.radius)/grid_size)*grid_size
		local x2 = math.floor ((particle.x+particle.radius)/grid_size)*grid_size
		local y1 = math.floor ((particle.y-particle.radius)/grid_size)*grid_size
		local y2 = math.floor ((particle.y+particle.radius)/grid_size)*grid_size
		
		for x = x1, x2, grid_size do
			for y = y1, y2, grid_size do
				if grid[x] and grid[x][y] then
					table.insert (grid[x][y].particles, particle)
				end	
			end
		end
	end
	
	collisions = {}
	for x,ys in pairs (grid) do
		for y, g in pairs (ys) do
			if #g.particles >= 2 then
				g.near = true
				if is_overlap (g.particles) then
					g.overlap = true
	--				pushback (g.particles)
				else
					g.overlap = false
				end
			else
				g.near = false
				g.overlap = false
			end

		end
	end
end

function love.update(dt)
	if dt > 0.1 then dt = 0.1 end
	for i, particle in pairs (particles) do
		update_particle (particle, dt)
	end
	update_grid (dt)
--	solve_grid (dt)
end


function love.draw()
--	print('draw')
	for x, ys in pairs (grid) do
		for y, g in pairs (ys) do
			if g.overlap then
				love.graphics.setColor(1,0,0)
			elseif g.near then
				love.graphics.setColor(1,1,0)
			elseif #g.particles > 0 then
				love.graphics.setColor(0.6,0.4,0)
			else
				love.graphics.setColor(g.color)
			end
			love.graphics.rectangle('fill', x, y, g.size, g.size)
		end
	end
	for i, particle in pairs (particles) do
		love.graphics.setColor(particle.color)
		love.graphics.circle('fill', particle.x, particle.y, particle.radius)
	end
	love.graphics.setColor(1,0,0)
	for i, js in pairs (collisions) do
		for j, collision in pairs (js) do
			local particles = collision.particles
			love.graphics.line(particles[1].x, particles[1].y, particles[2].x, particles[2].y)
		end
	end
end

function love.keypressed(key, scancode, isrepeat)
   if key == "escape" then
      love.event.quit()
   end
end